# -*- coding: utf-8 -*-

"""
这个模块可以生成一个对所有的 SdmMacro YAML 文件进行枚举的 Python 模块.
"""

from pathlib_mate import Path

from .model import SdmMacro


def slugify(s: str) -> str:
    """
    将字符串转换成一个合法的 Python 变量名.
    """
    return s.replace(" ", "_").replace("-", "_").replace("/", "__").replace("\\", "__")


def get_var_name(
    dir: Path,
    path: Path,
):
    """
    从文件路径生成一个合法的 Python 变量名作为 Enum 枚举值的变量名. ``path`` 是这个 yaml 文件
    的路径, 而 ``dir`` 则是我们在搜索 yaml 文件时的起始点根目录.
    例如我们有一个 ``${HOME}/sdm_macros/warrior/main_rotation.yml`` 文件, 而 ``dir`` 是
    ``${HOME}/sdm_macros/``. 那么这个模板文件的变量名就会是 ``warrior__main_rotation``.

    :param dir:
    :param path:
    """
    relpath = path.relative_to(dir)
    var_name = slugify(str(relpath)).split(".")[0]
    if var_name[0].isalpha() is False:  # 如果第一个字符不是字母, 那么加上一个 f_ (file)
        var_name = "f_" + var_name
    return var_name


def to_module(
    dir_root: Path,
    import_dir_root_line: str,
):
    """
    扫描 ``dir_root`` 文件夹下的所有宏命令 yaml 文件, 生成一个 Python 模块的字符串.
    这个模块包含了所有的宏命令 yaml 文件的枚举.

    那么最终生成的 Python 模块请参考 :ref:`generate-wtf-config-enum-module` todo: fix this ref

    :param dir_root: 宏命令 yaml 文件的根目录.
    :param import_dir_root_line: 这一行要导入一个 dir_root 对象, 也就是我们扫描的
        宏命令 yaml 文件的根目录. 用于里面的 enum 中的路径的拼接.
    """
    lines = [
        "# -*- coding: utf-8 -*-",
        "",
        import_dir_root_line,
        "",
        "# fmt: off",
        "class MacroEnum:",
    ]
    tab = " " * 4
    paths = list(Path.sort_by_abspath(dir_root.select_by_ext(".yml")))
    if len(paths):
        for path in paths:
            # make sure the macro id in file name match the yaml data
            macro = SdmMacro.from_yaml(path)
            parts = path.fname.split("-", 1)
            if str(macro.id) != parts[0]:  # pragma: no cover
                raise ValueError(f"Macro Id doesn't match file name in {path}")
            # figure out the variable name
            var_name = get_var_name(dir_root, path)
            relpath = path.relative_to(dir_root)
            joinpath_arg = ", ".join([f'"{part}"' for part in relpath.parts])
            lines.append(
                f"{tab}{var_name} = dir_root.joinpath({joinpath_arg}) # file://{path}"
            )
    else:  # pragma: no cover
        lines.append("    pass")
    lines.append("# fmt: on")
    lines.append("")
    return "\n".join(lines)



# -*- coding: utf-8 -*-

"""
为了能实现用 Python 来操作 SDM 插件的 ``SavedVariables/SuperDuperMacro.lua`` 文件,
我们要对 .lua 文件中的代码块进行数据建模, 定义一些 Python 类来代表这些代码块, 然后实现将
Python 对象转化为 .lua 代码块的逻辑.
"""

import typing as T
import dataclasses

from jinja2 import Template
from pathlib_mate import Path
from yaml import load, SafeLoader

from ..utils import apply


_dir_here = Path.dir_here(__file__)
_path_sdm_template = _dir_here / "sdm.tpl"
_sdm_template = Template(_path_sdm_template.read_text(encoding="utf-8"))


@dataclasses.dataclass
class SdmCharacter:
    """
    代表 Character Macro 专有宏命令中关于角色信息的部分. 对应着如下代码块:

    .. code-block:: lua

        ["character"] = {
            ["name"] = "charname",
            ["realm"] = "realmname",
        },
    """

    name: str = dataclasses.field()
    realm: str = dataclasses.field()


class SdmMacroTypeEnum:
    """
    枚举 SDM 宏命令的三种类型.
    """

    button = "b"
    floating = "f"
    script = "s"


_DEFAULT_TYPE = SdmMacroTypeEnum.button  # 默认的宏命令类型
_DEFAULT_ID = 0  # 默认的起始 ID
_DEFAULT_ICON = 1  # 默认的宏命令图标


@dataclasses.dataclass
class SdmMacro:
    """
    定义了一个魔兽世界中的 SDM 宏命令的抽象, 目前只支持 Button + Global 这一种模式.

    代表着如下代码块:

    .. code-block:: lua

        {
            ["type"] = "f",
            ["name"] = "macroname",
            ["character"] = {
                ["name"] = "charname",
                ["realm"] = "realmname",
            },
            ["ID"] = 1,
            ["text"] = "/s hello",
            ["icon"] = 1,
        }, -- [1]
    """

    name: str = dataclasses.field()
    character: T.Optional[SdmCharacter] = dataclasses.field(default=None)
    type: str = dataclasses.field(default=_DEFAULT_TYPE)
    id: int = dataclasses.field(default=_DEFAULT_ID)  # SDM macro ID starts from 0
    icon: int = dataclasses.field(default=_DEFAULT_ICON)  # 1 is the Question Mark Icon
    text: str = dataclasses.field(default="")

    def set_id(self, id: int) -> "SdmMacro":  # pragma: no cover
        """
        Update it's attributes value.
        """
        self.id = id
        return self

    def set_char(self, name: str, realm: str) -> "SdmMacro":  # pragma: no cover
        """
        Update it's attributes value.
        """
        if self.character is None:
            self.character = SdmCharacter(
                name=name,
                realm=realm,
            )
        else:
            self.character.name = name
            self.character.realm = realm
        return self

    def is_global(self) -> bool:
        """
        Is this SDM macro a global macro or character macro
        """
        if self.character is None:  # pragma: no cover
            return True
        elif (self.character.name is None) or (self.character.realm is None):
            return True
        else:
            return False

    def encode_text(self) -> str:
        """
        Encode macro text to single-ling Lua string. The final string of the
        macro body in lua has to have only one line.
        """
        return self.text.replace("\n", "\\n")

    @classmethod
    def from_yaml(cls, stream) -> "SdmMacro":
        """
        从人类可读写的 yaml 文件中读取数据, 创建 :class:`SDMMacro` 对象. 这是我们
        这个模块的最核心的方法, 也是能让我们用 yaml 文件来维护宏命令的关键.

        下面是一个示例的 yaml 文件.

        .. code-block:: yaml

            name: interrupt
            character:
              name:
              realm:
            type: b
            id:
            # you can find icon id on https://wotlk.evowow.com/?icons
            icon:
            description: |
              cancel casting spell, interrupt enemy casting immediately!
            text: |
              #showtooltip
              /stopcasting
              /cast Counterspell

        :param stream: 可以是 yaml 文件的字符串内容, 也可是 yaml 的 Path 对象,
            也可以是 file object 对象.
        """
        if isinstance(stream, str):  # pragma: no cover
            data = load(stream, SafeLoader)
        elif isinstance(stream, Path):
            with stream.open("r", encoding="utf-8") as f:
                data = load(f, SafeLoader)
        else:  # pragma: no cover
            data = load(stream, SafeLoader)
        return cls(
            name=data["name"],
            character=SdmCharacter(**data["character"]),
            type=data["type"] if data["type"] else _DEFAULT_TYPE,
            id=data["id"] if data["id"] else _DEFAULT_ID,
            icon=data["icon"] if data["icon"] else _DEFAULT_ICON,
            text=data["text"].strip(),
        )

    def render(self) -> str:
        """
        Render the corresponding SuperDupeMacro.lua code. See example at
        :class:`SDMMacro`.
        """
        lines: T.List[str] = list()
        lines.append(f"[{self.id}] = {{")
        lines.append(f'    ["type"] = "{self.type}",')
        lines.append(f'    ["name"] = "{self.name}",')
        if self.is_global() is False:
            lines.append(f'    ["character"] = {{')
            lines.append(f'        ["name"] = "{self.character.name}",')
            lines.append(f'        ["realm"] = "{self.character.realm}",')
            lines.append(f"    }},")
        lines.append(f'    ["ID"] = {self.id},')
        lines.append(f'    ["icon"] = {self.icon},')
        lines.append(f'    ["text"] = "{self.encode_text()}",')
        lines.append(f"}},")
        return "\n".join(lines)


@dataclasses.dataclass
class SdmLua:
    """
    代表了 ``SuperDupeMacro.lua`` 文件的抽象. 该类只能用于将数据写入到 ``SuperDupeMacro.lua``,
    而不能从 ``SuperDupeMacro.lua`` 中读取数据.

    :param path_lua: ``SuperDupeMacro.lua`` 文件路径.
    :param macros: :class:`SDMMacro` 对象的列表.
    """

    path_lua: Path = dataclasses.field()
    macros: T.List[SdmMacro] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        self.check_path_lua()
        self.check_macros()

    def check_path_lua(self):
        """
        检查 :attr:`SdmLua.path_lua` 是否是 SuperDupeMacro.lua 文件.
        """
        if self.path_lua.basename != "SuperDuperMacro.lua":  # pragma: no cover
            raise ValueError(f"the SDMLua.path_lua has to end with SuperDupeMacro.lua!")

    def check_macros(self):
        """
        检查 :attr:`SdmLua.macros` 中是否有重复的 macro id.
        """
        id_set = {macro.id for macro in self.macros}
        if len(id_set) != len(self.macros):  # pragma: no cover
            macro_id_list = [macro.id for macro in self.macros]
            raise ValueError(
                f"Cannot render SDM lua! Found duplicate id in 'macro_list': {macro_id_list}"
            )

    def render(self) -> str:
        """
        将一堆 :class:`SDMMacro` 对象渲染成 SuperDupeMacro.lua 文件的内容 (只是生成内容
        而不将内容写入文件). 这里面会检查 macro_list 中的 macro id 是否有重复, 如果有重复,
        则会抛出异常.
        """
        return _sdm_template.render(macros=self.macros)

    def write(
        self,
        real_run: bool = True,
        verbose: bool = False,
    ) -> str:  # pragma: no cover
        content = self.render()
        apply(
            path=self.path_lua,
            content=content,
            real_run=real_run,
            verbose=verbose,
        )
        return content



# -*- coding: utf-8 -*-

import typing as T
import itertools

from pathlib_mate import Path
from ordered_set import OrderedSet

from .logger import logger


def get_values(enum_class) -> OrderedSet:
    """
    Get all values from an enum class.

    Example::

        >>> class Color:
        ...     RED = 1
        ...     GREEN = 2
        ...     BLUE = 3
        >>> get_values(Color)
        OrderedSet([1, 2, 3])
    """
    return OrderedSet(
        [v for k, v in enum_class.__dict__.items() if not k.startswith("_")]
    )


KT = T.TypeVar("KT")
VT = T.TypeVar("VT")


def group_by(
    iterable: T.Iterable[VT],
    get_key: T.Callable[[VT], KT],
) -> T.Dict[KT, T.List[VT]]:  # pragma: no cover
    """
    Group items by it's key, with type hint.

    Example::

        >>> class Record:
        ...     def __init__(self, product: str, date: str, sale: int):
        ...         self.product = product
        ...         self.date = date
        ...         self.sale = sale

        >>> records = [
        ...     Record("apple", "2020-01-01", 10),
        ...     Record("apple", "2020-01-02", 20),
        ...     Record("apple", "2020-01-03", 30),
        ...     Record("banana", "2020-01-01", 10),
        ...     Record("banana", "2020-01-02", 20),
        ...     Record("banana", "2020-01-03", 30),
        ... ]

        >>> group_by(records, lambda x: x.product)
        {
            "apple": [
                Record("apple", "2020-01-01", 10),
                Record("apple", "2020-01-02", 20),
                Record("apple", "2020-01-03", 30),
            ],
            "banana": [
                Record("banana", "2020-01-01", 10),
                Record("banana", "2020-01-02", 20),
                Record("banana", "2020-01-03", 30),
            ],
        }
    """
    grouped = dict()
    for item in iterable:
        key = get_key(item)
        try:
            grouped[key].append(item)
        except KeyError:
            grouped[key] = [item]
    return grouped


def concat_lists(*lists) -> list:
    """
    Concatenate multiple lists into one list.

    Example::

        >>> concat_lists([1, 2], [3, 4], [5, 6])
        [1, 2, 3, 4, 5, 6]
    """
    return list(itertools.chain(*lists))


def apply(
    path: Path,
    content: str,
    real_run: bool = False,
    verbose: bool = True,
):
    r"""
    Apply a content to a file (Write to the file in ``World of Warcraft\WTF\...``).

    :param path: The file path to write to.
    :param content: The content to write.
    :param real_run: If True, do not write to the file.
    :param verbose: If True, print log.
    """
    if verbose:
        # 使用 file:// URI 使得可以点击打印在 Console 中的输出直接跳转到文件
        logger.info(f"Write to: file://{path}")
    if real_run:
        try:
            path.write_text(content, encoding="utf-8")
        except FileNotFoundError:
            path.parent.mkdir(parents=True)
            path.write_text(content, encoding="utf-8")



# -*- coding: utf-8 -*-

from wow_sdm import api


def test():
    _ = api
    _ = api.get_values
    _ = api.group_by
    _ = api.concat_lists
    _ = api.logger
    _ = api.exp03_wotlk

    exp03_wotlk = api.exp03_wotlk
    _ = exp03_wotlk.SdmCharacter
    _ = exp03_wotlk.SdmMacroTypeEnum
    _ = exp03_wotlk.SdmMacro
    _ = exp03_wotlk.SdmLua
    _ = exp03_wotlk.Client
    _ = exp03_wotlk.AccLvlMapping
    _ = exp03_wotlk.CharLvlMapping
    _ = exp03_wotlk.SdmMapping
    _ = exp03_wotlk.to_module
    _ = exp03_wotlk.Client.get_account_sdm_lua
    _ = exp03_wotlk.AccLvlMapping.make_many
    _ = exp03_wotlk.CharLvlMapping.make_many
    _ = exp03_wotlk.SdmMapping.apply


if __name__ == "__main__":
    from wow_sdm.tests import run_cov_test

    run_cov_test(__file__, "wow_sdm.api", preview=False)



# -*- coding: utf-8 -*-

"""
这个模块的可以让你对 Account / Character 和宏命令 Yaml 文件进行排列组合, 然后一键将你的
配置文件应用到你的客户端中的 WTF 目录下.
"""

import typing as T
import dataclasses
from pathlib_mate import Path

from wow_acc.api import Account, Character

from ..logger import logger

from .model import SdmMacro, SdmLua


@dataclasses.dataclass
class Client:
    r"""
    代表着一个具体魔兽世界客户端. 你必须要指定这个客户端的目录. 然后就可以用各种 method 来
    获取对应的 SDM 插件 lua 文件的绝对路径了.

    :param dir: 客户端目录, 根据此目录可以定位其他的目录. 例如 "C:\Program Files\World of Warcraft"
    :param locale: 客户端语种, 例如 enUS, zhCN, zhTW 等.
    """

    dir: Path = dataclasses.field()
    locale: str = dataclasses.field()

    @property
    def dir_wtf(self) -> Path:
        return self.dir / "WTF"

    def get_account_sdm_lua(
        self,
        wtf_account_name: str,
    ) -> Path:
        r"""
        This file stores account level SDM AddOn configuration.

        Example: ``C:\...\WTF\Account\MYACCOUNT\SavedVariables\SuperDuperMacro.lua``
        """
        return self.dir_wtf.joinpath(
            "Account",
            wtf_account_name,
            "SavedVariables",
            "SuperDuperMacro.lua",
        )


@dataclasses.dataclass
class BaseMapping:
    """
    Mapping 是一个 Account / Character 和一个 Macro 文件的组合.
    """

    file: Path = dataclasses.field()


@dataclasses.dataclass
class AccLvlMapping(BaseMapping):
    """
    ``Account`` 和一个 Macro 文件的组合.

    :param acc: `wow_acc.api.Account <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Account>`_ 对象.
    :param file: Macro 文件的绝对路径.
    """

    acc: Account = dataclasses.field()
    file: Path = dataclasses.field()

    @classmethod
    def make_many(
        cls,
        account: T.Union[Account, T.Iterable[Account]],
        file: T.Union[Path, T.Iterable[Path]],
    ):
        """
        生成多个 ``Account`` 和 Macro 文件的组合. 让你写更少的代码.

        :param account: 单个 Account 或是多个 Account 的集合.
        :param file: 单个 Path 或是多个 Path 的集合.
        """
        if isinstance(account, Account):  # pragma: no cover
            accounts = [account]
        else:  # pragma: no cover
            accounts = account
        if isinstance(file, Path):  # pragma: no cover
            files = [file]
        else:
            files = file

        lst = list()
        for account in accounts:
            for file in files:
                lst.append(cls(acc=account, file=file))
        return lst


@dataclasses.dataclass
class CharLvlMapping(BaseMapping):
    """
    ``Character`` 和一个 Macro 文件的组合.

    :param char: `wow_acc.api.Character <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Character>`_ 对象.
    :param file: Macro 文件的绝对路径.
    """

    char: Character = dataclasses.field()
    file: Path = dataclasses.field()

    @classmethod
    def make_many(
        cls,
        character: T.Union[Character, T.Iterable[Character]],
        file: T.Union[Path, T.List[Path]],
    ):
        """
        生成多个 ``Character`` 和 Macro 文件的组合. 让你写更少的代码.

        :param character: 单个 Character 或是多个 Character 的集合.
        :param file: 单个 Path 或是多个 Path 的集合.
        """
        if isinstance(character, Character):  # pragma: no cover
            characters = [character]
        else:  # pragma: no cover
            characters = character
        if isinstance(file, Path):  # pragma: no cover
            files = [file]
        else:
            files = file

        lst = list()
        for character in characters:
            for file in files:
                lst.append(cls(char=character, file=file))
        return lst


@dataclasses.dataclass
class SdmMapping:
    """
    定义了一个魔兽世界客户端中被管理的所有 Sdm 宏命令的设定.

    :param client: :class:`Client` 对象. 有了这个才知道我们要将配置文件写到哪里去.
    :param all_accounts: 所有的 `wow_acc.api.Account <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Account>`_ 对象.
        在 render Jinja 模板时会用到.
    :param all_characters: 所有的 `wow_acc.api.Character <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Character>`_ 对象.
        在 render Jinja 模板时会用到.
    """

    # fmt: off
    client: Client
    all_accounts: T.Iterable[Account]
    all_characters: T.Iterable[Character]

    acc_macros: T.List[AccLvlMapping] = dataclasses.field(default_factory=list)
    char_macros: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    # fmt: on

    @logger.emoji_block(msg="{func_name}", emoji="🎮")
    def apply(self, real_run: bool = False):
        """
        将 SDM 的 Lua 文件写入 ``SavedVariables`` 文件夹中.
        """
        # mapper 是一个以 account 为 key 的分组器
        mapper: T.Dict[str, T.List[SdmMacro]] = dict()

        for acc_map in self.acc_macros:
            macro = SdmMacro.from_yaml(acc_map.file)
            try:
                mapper[acc_map.acc.wtf_account_name].append(macro)
            except KeyError:
                mapper[acc_map.acc.wtf_account_name] = [macro]

        for char_map in self.char_macros:
            macro = SdmMacro.from_yaml(char_map.file)
            # 将其设定为 character macro
            macro.set_char(name=char_map.char.titled_character_name, realm=char_map.char.realm_name)
            try:
                mapper[char_map.char.account.wtf_account_name].append(macro)
            except KeyError:  # pragma: no cover
                mapper[char_map.char.account.wtf_account_name] = [macro]

        for wtf_account_name, macros in mapper.items():
            path_lua = self.client.get_account_sdm_lua(wtf_account_name)
            sdm_lua = SdmLua(
                path_lua=path_lua,
                macros=macros,
            )
            sdm_lua.write(real_run=real_run, verbose=True)



# -*- coding: utf-8 -*-

from .logger import logger
from .utils import get_values
from .utils import group_by
from .utils import concat_lists
from .utils import apply
from .exp03_wotlk import api as exp03_wotlk
